@@ -20,17 +20,29 @@ from __future__ import annotations
 
 import copy
 import traceback
+from collections import Counter
 from collections.abc import Mapping
 from dataclasses import dataclass
-from typing import Dict, Generic, TypeVar, cast, overload
+from pathlib import Path
+from typing import ClassVar, Dict, Generic, MutableMapping, Optional, TypeVar, cast, overload
 
+import absl
+import dill
 import numpy as np
 import tensorflow as tf
-from absl import logging
 
+from .acquisition.multi_objective import non_dominated
+
+try:
+    import pandas as pd
+    import seaborn as sns
+except ModuleNotFoundError:
+    pd = None
+    sns = None
+
+from . import logging
 from .acquisition.rule import AcquisitionRule, EfficientGlobalOptimization
 from .data import Dataset
-from .logging import get_tensorboard_writer, set_step_number
 from .models import ModelSpec, TrainableProbabilisticModel, create_model
 from .models.config import ModelConfigType
 from .observer import OBJECTIVE, Observer
@@ -77,7 +89,57 @@ class Record(Generic[StateType]):
         if len(self.models) == 1:
             return next(iter(self.models.values()))
         else:
-            raise ValueError(f"Expected a single dataset, found {len(self.datasets)}")
+            raise ValueError(f"Expected a single model, found {len(self.models)}")
+
+    def save(self, path: Path | str) -> FrozenRecord[StateType]:
+        """Save the record to disk. Will overwrite any existing file at the same path."""
+        Path(path).parent.mkdir(exist_ok=True, parents=True)
+        with open(path, "wb") as f:
+            dill.dump(self, f, dill.HIGHEST_PROTOCOL)
+        return FrozenRecord(Path(path))
+
+
+@dataclass(frozen=True)
+class FrozenRecord(Generic[StateType]):
+    """
+    A Record container saved on disk.
+
+    Note that records are saved via pickling and are therefore neither portable nor secure.
+    Only open frozen records generated on the same system.
+    """
+
+    path: Path
+    """ The path to the pickled Record. """
+
+    def load(self) -> Record[StateType]:
+        """Load the record into memory."""
+        with open(self.path, "rb") as f:
+            return dill.load(f)
+
+    @property
+    def datasets(self) -> Mapping[str, Dataset]:
+        """The known data from the observer."""
+        return self.load().datasets
+
+    @property
+    def models(self) -> Mapping[str, TrainableProbabilisticModel]:
+        """The models over the :attr:`datasets`."""
+        return self.load().models
+
+    @property
+    def acquisition_state(self) -> StateType | None:
+        """The acquisition state."""
+        return self.load().acquisition_state
+
+    @property
+    def dataset(self) -> Dataset:
+        """The dataset when there is just one dataset."""
+        return self.load().dataset
+
+    @property
+    def model(self) -> TrainableProbabilisticModel:
+        """The model when there is just one dataset."""
+        return self.load().model
 
 
 # this should be a generic NamedTuple, but mypy doesn't support them
@@ -92,14 +154,24 @@ class OptimizationResult(Generic[StateType]):
     exception.
     """
 
-    history: list[Record[StateType]]
+    history: list[Record[StateType] | FrozenRecord[StateType]]
     r"""
     The history of the :class:`Record`\ s from each step of the optimization process. These
-    :class:`Record`\ s are created at the *start* of each loop, and as such will never include the
-    :attr:`final_result`.
+    :class:`Record`\ s are created at the *start* of each loop, and as such will never
+    include the :attr:`final_result`. The records may be either in memory or on disk.
     """
 
-    def astuple(self) -> tuple[Result[Record[StateType]], list[Record[StateType]]]:
+    @staticmethod
+    def step_filename(step: int, num_steps: int) -> str:
+        """Default filename for saved optimization steps."""
+        return f"step.{step:0{len(str(num_steps - 1))}d}.pickle"
+
+    STEP_GLOB: ClassVar[str] = "step.*.pickle"
+    RESULTS_FILENAME: ClassVar[str] = "results.pickle"
+
+    def astuple(
+        self,
+    ) -> tuple[Result[Record[StateType]], list[Record[StateType] | FrozenRecord[StateType]]]:
         """
         **Note:** In contrast to the standard library function :func:`dataclasses.astuple`, this
         method does *not* deepcopy instance attributes.
@@ -131,6 +203,19 @@ class OptimizationResult(Generic[StateType]):
         else:
             raise ValueError(f"Expected a single dataset, found {len(datasets)}")
 
+    def try_get_optimal_point(self) -> tuple[TensorType, TensorType, TensorType]:
+        """
+        Convenience method to attempt to get the optimal point for a single dataset,
+        single objective run.
+
+        :return: Tuple of the optimal query point, observation and its index.
+        """
+        dataset = self.try_get_final_dataset()
+        if tf.rank(dataset.observations) != 2 or dataset.observations.shape[1] != 1:
+            raise ValueError("Expected a single objective")
+        arg_min_idx = tf.squeeze(tf.argmin(dataset.observations, axis=0))
+        return dataset.query_points[arg_min_idx], dataset.observations[arg_min_idx], arg_min_idx
+
     def try_get_final_models(self) -> Mapping[str, TrainableProbabilisticModel]:
         """
         Convenience method to attempt to get the final models.
@@ -154,6 +239,40 @@ class OptimizationResult(Generic[StateType]):
         else:
             raise ValueError(f"Expected single model, found {len(models)}")
 
+    @property
+    def loaded_history(self) -> list[Record[StateType]]:
+        """The history of the optimization process loaded into memory."""
+        return [record if isinstance(record, Record) else record.load() for record in self.history]
+
+    def save_result(self, path: Path | str) -> None:
+        """Save the final result to disk. Will overwrite any existing file at the same path."""
+        Path(path).parent.mkdir(exist_ok=True, parents=True)
+        with open(path, "wb") as f:
+            dill.dump(self.final_result, f, dill.HIGHEST_PROTOCOL)
+
+    def save(self, base_path: Path | str) -> None:
+        """Save the optimization result to disk. Will overwrite existing files at the same path."""
+        path = Path(base_path)
+        num_steps = len(self.history)
+        self.save_result(path / self.RESULTS_FILENAME)
+        for i, record in enumerate(self.loaded_history):
+            record_path = path / self.step_filename(i, num_steps)
+            record.save(record_path)
+
+    @classmethod
+    def from_path(cls, base_path: Path | str) -> OptimizationResult[StateType]:
+        """Load a previously saved OptimizationResult."""
+        try:
+            with open(Path(base_path) / cls.RESULTS_FILENAME, "rb") as f:
+                result = dill.load(f)
+        except FileNotFoundError as e:
+            result = Err(e)
+
+        history: list[Record[StateType] | FrozenRecord[StateType]] = [
+            FrozenRecord(file) for file in sorted(Path(base_path).glob(cls.STEP_GLOB))
+        ]
+        return cls(result, history)
+
 
 class BayesianOptimizer(Generic[SearchSpaceType]):
     """
@@ -183,6 +302,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         model_specs: Mapping[str, ModelSpec],
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[None]:
         ...
@@ -198,6 +318,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         ],
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
         # this should really be OptimizationResult[None], but tf.Tensor is untyped so the type
         # checker can't differentiate between TensorType and State[S | None, TensorType], and
@@ -217,6 +338,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         ],
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
         # this should really be OptimizationResult[None], but tf.Tensor is untyped so the type
         # checker can't differentiate between TensorType and State[S | None, TensorType], and
@@ -236,6 +358,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         acquisition_state: StateType | None = None,
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[StateType]:
         ...
@@ -253,6 +376,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         acquisition_state: StateType | None = None,
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[StateType]:
         ...
@@ -265,6 +389,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         model_specs: ModelSpec,
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[None]:
         ...
@@ -280,6 +405,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         ],
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[object]:
         ...
@@ -295,6 +421,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         ],
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[object]:
         ...
@@ -311,6 +438,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         acquisition_state: StateType | None = None,
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[StateType]:
         ...
@@ -327,6 +455,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         acquisition_state: StateType | None = None,
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[StateType]:
         ...
@@ -348,6 +477,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
         acquisition_state: StateType | None = None,
         *,
         track_state: bool = True,
+        track_path: Optional[Path | str] = None,
         fit_initial_model: bool = True,
     ) -> OptimizationResult[StateType] | OptimizationResult[None]:
         """
@@ -362,14 +492,10 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
             - Updates the datasets and models with the data from the ``observer``.
 
         If any errors are raised during the optimization loop, this method will catch and return
-        them instead, along with the history of the optimization process, and print a message (using
-        `absl` at level `logging.ERROR`).
-
-        **Note:** While the :class:`~trieste.models.TrainableProbabilisticModel` interface implies
-        mutable models, it is *not* guaranteed that the model passed to :meth:`optimize` will
-        be updated during the optimization process. For example, if ``track_state`` is `True`, a
-        copied model will be used on each optimization step. Use the models in the return value for
-        reliable access to the updated models.
+        them instead and print a message (using `absl` at level `absl.logging.ERROR`).
+        If ``track_state`` is enabled, then in addition to the final result, the history of the
+        optimization process will also be returned. If ``track_path`` is also set, then
+        the history and final result will be saved to disk rather than all being kept in memory.
 
         **Type hints:**
             - The ``acquisition_rule`` must use the same type of
@@ -392,6 +518,8 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
             :class:`Record`.
         :param track_state: If `True`, this method saves the optimization state at the start of each
             step. Models and acquisition state are copied using `copy.deepcopy`.
+        :param track_path: If set, the optimization state is saved to disk at this path,
+            rather than being copied in memory.
         :param fit_initial_model: If `False`, this method assumes that the initial models have
             already been optimized on the datasets and so do not require optimization before the
             first optimization step.
@@ -445,20 +573,48 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
             Dict[str, TrainableProbabilisticModelType], map_values(create_model, model_specs)
         )
 
-        history: list[Record[StateType]] = []
+        history: list[FrozenRecord[StateType] | Record[StateType]] = []
+        query_plot_dfs: dict[int, pd.DataFrame] = {}
+        observation_plot_dfs: dict[str, pd.DataFrame] = {}
+
+        summary_writer = logging.get_tensorboard_writer()
+        if summary_writer:
+            with summary_writer.as_default(step=0):
+                self._write_summary_init(
+                    acquisition_rule, datasets, models, num_steps, observation_plot_dfs
+                )
 
         for step in range(num_steps):
-            set_step_number(step)
+            logging.set_step_number(step)
             try:
 
                 if track_state:
-                    models_copy = copy.deepcopy(models)
-                    acquisition_state_copy = copy.deepcopy(acquisition_state)
-                    history.append(Record(datasets, models_copy, acquisition_state_copy))
+                    try:
+                        if track_path is None:
+                            datasets_copy = copy.deepcopy(datasets)
+                            models_copy = copy.deepcopy(models)
+                            acquisition_state_copy = copy.deepcopy(acquisition_state)
+                            record = Record(datasets_copy, models_copy, acquisition_state_copy)
+                            history.append(record)
+                        else:
+                            track_path = Path(track_path)
+                            record = Record(datasets, models, acquisition_state)
+                            file_name = OptimizationResult.step_filename(step, num_steps)
+                            history.append(record.save(track_path / file_name))
+                    except Exception as e:
+                        raise NotImplementedError(
+                            "Failed to save the optimization state. Some models do not support "
+                            "deecopying or serialization and cannot be saved. "
+                            "(This is particularly common for deep neural network models.) "
+                            "For these models, the `track_state`` argument of the "
+                            ":meth:`~trieste.bayesian_optimizer.BayesianOptimizer.optimize` method "
+                            "should be set to `False`. This means that only the final model "
+                            "will be available."
+                        ) from e
 
                 with Timer() as total_step_wallclock_timer:
-                    if step == 0 and fit_initial_model:
-                        with Timer() as initial_model_fitting_timer:
+                    with Timer() as initial_model_fitting_timer:
+                        if step == 0 and fit_initial_model:
                             for tag, model in models.items():
                                 dataset = datasets[tag]
                                 model.update(dataset)
@@ -468,11 +624,10 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
                         points_or_stateful = acquisition_rule.acquire(
                             self._search_space, models, datasets=datasets
                         )
-
-                    if callable(points_or_stateful):
-                        acquisition_state, query_points = points_or_stateful(acquisition_state)
-                    else:
-                        query_points = points_or_stateful
+                        if callable(points_or_stateful):
+                            acquisition_state, query_points = points_or_stateful(acquisition_state)
+                        else:
+                            query_points = points_or_stateful
 
                     observer_output = self._observer(query_points)
 
@@ -490,41 +645,19 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
                             model.update(dataset)
                             model.optimize(dataset)
 
-                summary_writer = get_tensorboard_writer()
                 if summary_writer:
-                    with summary_writer.as_default():
-                        for tag in datasets:
-                            with tf.name_scope(f"{tag}.model"):
-                                models[tag].log()
-                            tf.summary.scalar(
-                                f"{tag}.observation.best_overall",
-                                np.min(datasets[tag].observations),
-                                step=step,
-                            )
-                            tf.summary.scalar(
-                                f"{tag}.observation.best_new",
-                                np.min(tagged_output[tag].observations),
-                                step=step,
-                            )
-                        tf.summary.scalar(
-                            "wallclock.step",
-                            total_step_wallclock_timer.time,
-                            step=step,
-                        )
-                        tf.summary.scalar(
-                            "wallclock.query_point_generation",
-                            query_point_generation_timer.time,
-                            step=step,
-                        )
-                        tf.summary.scalar(
-                            "wallclock.model_fitting",
-                            model_fitting_timer.time
-                            + (
-                                initial_model_fitting_timer.time
-                                if (step == 0 and fit_initial_model)
-                                else 0
-                            ),
-                            step=step,
+                    with summary_writer.as_default(step=step):
+                        self._write_summary_step(
+                            datasets,
+                            models,
+                            query_points,
+                            tagged_output,
+                            initial_model_fitting_timer,
+                            model_fitting_timer,
+                            query_point_generation_timer,
+                            total_step_wallclock_timer,
+                            observation_plot_dfs,
+                            query_plot_dfs,
                         )
 
             except Exception as error:  # pylint: disable=broad-except
@@ -534,7 +667,7 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
                     f"\nTerminating optimization and returning the optimization history. You may "
                     f"be able to use the history to restart the process from a previous successful "
                     f"optimization step.\n",
-                    output_stream=logging.ERROR,
+                    output_stream=absl.logging.ERROR,
                 )
                 if isinstance(error, MemoryError):
                     tf.print(
@@ -542,11 +675,191 @@ class BayesianOptimizer(Generic[SearchSpaceType]):
                         "\nfunctions over large datasets, e.g. when initializing optimizers. "
                         "\nYou may be able to word around this by splitting up the evaluation "
                         "\nusing split_acquisition_function or split_acquisition_function_calls.",
-                        output_stream=logging.ERROR,
+                        output_stream=absl.logging.ERROR,
                     )
-                return OptimizationResult(Err(error), history)
+                result = OptimizationResult(Err(error), history)
+                if track_state and track_path is not None:
+                    result.save_result(Path(track_path) / OptimizationResult.RESULTS_FILENAME)
+                return result
 
-        tf.print("Optimization completed without errors", output_stream=logging.INFO)
+        tf.print("Optimization completed without errors", output_stream=absl.logging.INFO)
 
         record = Record(datasets, models, acquisition_state)
-        return OptimizationResult(Ok(record), history)
+        result = OptimizationResult(Ok(record), history)
+        if track_state and track_path is not None:
+            result.save_result(Path(track_path) / OptimizationResult.RESULTS_FILENAME)
+        return result
+
+    def _write_summary_init(
+        self,
+        acquisition_rule: AcquisitionRule[
+            TensorType | State[StateType | None, TensorType],
+            SearchSpaceType,
+            TrainableProbabilisticModelType,
+        ],
+        datasets: Mapping[str, Dataset],
+        models: Mapping[str, TrainableProbabilisticModel],
+        num_steps: int,
+        observation_plot_dfs: MutableMapping[str, pd.DataFrame],
+    ) -> None:
+        """Write initial TensorBoard summary (and set up any initial monitoring state)."""
+        devices = tf.config.list_logical_devices()
+        logging.text(
+            "metadata",
+            f"Observer: `{self._observer}`\n\n"
+            f"Number of steps: `{num_steps}`\n\n"
+            f"Number of initial points: "
+            f"`{dict((k, len(v)) for k, v in datasets.items())}`\n\n"
+            f"Search Space: `{self._search_space}`\n\n"
+            f"Acquisition rule:\n\n    {acquisition_rule}\n\n"
+            f"Models:\n\n    {models}\n\n"
+            f"Available devices: `{dict(Counter(d.device_type for d in devices))}`",
+        )
+
+        seaborn_warning = False
+        if logging.include_summary("query_points/_pairplot") and not (pd and sns):
+            seaborn_warning = True
+        for tag in datasets:
+            if logging.include_summary(f"{tag}.observations/_pairplot"):
+                output_dim = tf.shape(datasets[tag].observations)[-1]
+                if output_dim >= 2:
+                    if not (pd and sns):
+                        seaborn_warning = True
+                    else:
+                        columns = [f"x{i}" for i in range(output_dim)]
+                        observation_plot_dfs[tag] = pd.DataFrame(
+                            datasets[tag].observations, columns=columns
+                        ).applymap(float)
+                        observation_plot_dfs[tag]["observations"] = "initial"
+
+        if seaborn_warning:
+            tf.print(
+                "\nPairplot TensorBoard summaries require seaborn to be installed."
+                "\nOne way to do this is to install 'trieste[plotting]'.",
+                output_stream=absl.logging.INFO,
+            )
+
+    def _write_summary_step(
+        self,
+        datasets: Mapping[str, Dataset],
+        models: Mapping[str, TrainableProbabilisticModel],
+        query_points: TensorType,
+        tagged_output: Mapping[str, TensorType],
+        initial_model_fitting_timer: Timer,
+        model_fitting_timer: Timer,
+        query_point_generation_timer: Timer,
+        total_step_wallclock_timer: Timer,
+        observation_plot_dfs: MutableMapping[str, pd.DataFrame],
+        query_plot_dfs: MutableMapping[int, pd.DataFrame],
+    ) -> None:
+        """Write TensorBoard summary for the current step."""
+        for tag in datasets:
+
+            with tf.name_scope(f"{tag}.model"):
+                models[tag].log()
+
+            output_dim = tf.shape(tagged_output[tag].observations)[-1]
+            for i in tf.range(output_dim):
+                suffix = f"[{i}]" if output_dim > 1 else ""
+                if tf.size(tagged_output[tag].observations) > 0:
+                    logging.histogram(
+                        f"{tag}.observation{suffix}/new_observations",
+                        tagged_output[tag].observations[..., i],
+                    )
+                    logging.scalar(
+                        f"{tag}.observation{suffix}/best_new_observation",
+                        np.min(tagged_output[tag].observations[..., i]),
+                    )
+                if tf.size(datasets[tag].observations) > 0:
+                    logging.scalar(
+                        f"{tag}.observation{suffix}/best_overall",
+                        np.min(datasets[tag].observations[..., i]),
+                    )
+
+            if logging.include_summary(f"{tag}.observations/_pairplot") and (
+                pd and sns and output_dim >= 2
+            ):
+                columns = [f"x{i}" for i in range(output_dim)]
+                observation_new_df = pd.DataFrame(
+                    tagged_output[tag].observations, columns=columns
+                ).applymap(float)
+                observation_new_df["observations"] = "new"
+                observation_plot_df = pd.concat(
+                    (observation_plot_dfs.get(tag), observation_new_df),
+                    copy=False,
+                    ignore_index=True,
+                )
+                observation_plot_df["pareto"] = non_dominated(datasets[tag].observations)[1] == 0
+                observation_plot_df["observation type"] = observation_plot_df.apply(
+                    lambda x: x["observations"] + x["pareto"] * " (non-dominated)",
+                    axis=1,
+                )
+                pairplot = sns.pairplot(
+                    observation_plot_df,
+                    vars=columns,
+                    hue="observation type",
+                    hue_order=[
+                        "initial",
+                        "old",
+                        "new",
+                        "initial (non-dominated)",
+                        "old (non-dominated)",
+                        "new (non-dominated)",
+                    ],
+                    palette={
+                        "initial": "tab:green",
+                        "old": "tab:green",
+                        "new": "tab:orange",
+                        "initial (non-dominated)": "tab:purple",
+                        "old (non-dominated)": "tab:purple",
+                        "new (non-dominated)": "tab:red",
+                    },
+                    markers={
+                        "initial": "X",
+                        "old": "o",
+                        "new": "o",
+                        "initial (non-dominated)": "X",
+                        "old (non-dominated)": "o",
+                        "new (non-dominated)": "o",
+                    },
+                )
+                logging.pyplot(f"{tag}.observations/_pairplot", pairplot.fig)
+                observation_plot_df.loc[
+                    observation_plot_df["observations"] == "new", "observations"
+                ] = "old"
+                observation_plot_dfs[tag] = observation_plot_df
+
+        if tf.rank(query_points) == 2:
+            for i in tf.range(tf.shape(query_points)[1]):
+                if len(query_points) == 1:
+                    logging.scalar(f"query_points/[{i}]", float(query_points[0, i]))
+                else:
+                    logging.histogram(f"query_points/[{i}]", query_points[:, i])
+
+        if pd and sns and logging.include_summary("query_points/_pairplot"):
+            columns = [f"x{i}" for i in range(tf.shape(query_points)[1])]
+            query_new_df = pd.DataFrame(query_points, columns=columns).applymap(float)
+            query_new_df["query points"] = "new"
+            query_plot_df = pd.concat(
+                (query_plot_dfs.get(0), query_new_df), copy=False, ignore_index=True
+            )
+            pairplot = sns.pairplot(query_plot_df, hue="query points")
+            padding = 0.025 * (self._search_space.upper - self._search_space.lower)
+            upper_limits = self._search_space.upper + padding
+            lower_limits = self._search_space.lower - padding
+            for i in range(self._search_space.dimension):
+                pairplot.axes[0, i].set_xlim((lower_limits[i], upper_limits[i]))
+                pairplot.axes[i, 0].set_ylim((lower_limits[i], upper_limits[i]))
+            logging.pyplot("query_points/_pairplot", pairplot.fig)
+            query_plot_df["query points"] = "old"
+            query_plot_dfs[0] = query_plot_df
+
+        logging.scalar("wallclock/step", total_step_wallclock_timer.time)
+        logging.scalar(
+            "wallclock/query_point_generation",
+            query_point_generation_timer.time,
+        )
+        logging.scalar(
+            "wallclock/model_fitting",
+            model_fitting_timer.time + initial_model_fitting_timer.time,
+        )
