@@ -35,9 +35,10 @@ class SearchSpace(ABC):
     """
 
     @abstractmethod
-    def sample(self, num_samples: int) -> TensorType:
+    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
         """
         :param num_samples: The number of points to sample from this search space.
+        :param seed: Random seed for reproducibility.
         :return: ``num_samples`` i.i.d. random points, sampled uniformly from this search space.
         """
 
@@ -53,7 +54,7 @@ class SearchSpace(ABC):
 
     @property
     @abstractmethod
-    def dimension(self) -> int:
+    def dimension(self) -> TensorType:
         """The number of inputs in this search space."""
 
     @property
@@ -67,12 +68,33 @@ class SearchSpace(ABC):
         """The highest value taken by each search space dimension."""
 
     @abstractmethod
-    def __mul__(self: SearchSpaceType, other: SearchSpaceType) -> SearchSpaceType:
+    def product(self: SearchSpaceType, other: SearchSpaceType) -> SearchSpaceType:
         """
         :param other: A search space of the same type as this search space.
         :return: The Cartesian product of this search space with the ``other``.
         """
 
+    @overload
+    def __mul__(self: SearchSpaceType, other: SearchSpaceType) -> SearchSpaceType:
+        ...
+
+    @overload
+    def __mul__(self: SearchSpaceType, other: SearchSpace) -> SearchSpace:  # type: ignore[misc]
+        # mypy complains that this is superfluous, but it seems to use it fine to infer
+        # that Box * Box = Box, while Box * Discrete = SearchSpace.
+        ...
+
+    def __mul__(self, other: SearchSpace) -> SearchSpace:
+        """
+        :param other: A search space.
+        :return: The Cartesian product of this search space with the ``other``.
+            If both spaces are of the same type then this calls the :meth:`product` method.
+            Otherwise, it generates a :class:`TaggedProductSearchSpace`.
+        """
+        if isinstance(other, type(self)):
+            return self.product(other)
+        return TaggedProductSearchSpace((self, other))
+
     def __pow__(self: SearchSpaceType, other: int) -> SearchSpaceType:
         """
         Return the Cartesian product of ``other`` instances of this search space. For example, for
@@ -87,6 +109,21 @@ class SearchSpace(ABC):
         tf.debugging.assert_positive(other, message="Exponent must be strictly positive")
         return reduce(operator.mul, [self] * other)
 
+    def discretize(self, num_samples: int) -> DiscreteSearchSpace:
+        """
+        :param num_samples: The number of points in the :class:`DiscreteSearchSpace`.
+        :return: A discrete search space consisting of ``num_samples`` points sampled uniformly from
+            this search space.
+        """
+        return DiscreteSearchSpace(points=self.sample(num_samples))
+
+    @abstractmethod
+    def __eq__(self, other: object) -> bool:
+        """
+        :param other: A search space.
+        :return: Whether the search space is identical to this one.
+        """
+
 
 class DiscreteSearchSpace(SearchSpace):
     r"""
@@ -132,7 +169,7 @@ class DiscreteSearchSpace(SearchSpace):
         return self._points
 
     @property
-    def dimension(self) -> int:
+    def dimension(self) -> TensorType:
         """The number of inputs in this search space."""
         return self._dimension
 
@@ -140,21 +177,25 @@ class DiscreteSearchSpace(SearchSpace):
         tf.debugging.assert_shapes([(value, self.points.shape[1:])])
         return tf.reduce_any(tf.reduce_all(value == self._points, axis=1))
 
-    def sample(self, num_samples: int) -> TensorType:
+    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
         """
         :param num_samples: The number of points to sample from this search space.
+        :param seed: Random seed for reproducibility.
         :return: ``num_samples`` i.i.d. random points, sampled uniformly,
             from this search space.
         """
+        if seed is not None:  # ensure reproducibility
+            tf.random.set_seed(seed)
+
         if num_samples == 0:
             return self.points[:0, :]
         else:
             sampled_indices = tf.random.categorical(
-                tf.ones((1, tf.shape(self.points)[0])), num_samples
+                tf.ones((1, tf.shape(self.points)[0])), num_samples, seed=seed
             )
             return tf.gather(self.points, sampled_indices)[0, :, :]  # [num_samples, D]
 
-    def __mul__(self, other: DiscreteSearchSpace) -> DiscreteSearchSpace:
+    def product(self, other: DiscreteSearchSpace) -> DiscreteSearchSpace:
         r"""
         Return the Cartesian product of the two :class:`DiscreteSearchSpace`\ s. For example:
 
@@ -181,6 +222,15 @@ class DiscreteSearchSpace(SearchSpace):
         product_space_dimension = self.points.shape[-1] + other.points.shape[-1]
         return DiscreteSearchSpace(tf.reshape(cartesian_product, [-1, product_space_dimension]))
 
+    def __eq__(self, other: object) -> bool:
+        """
+        :param other: A search space.
+        :return: Whether the search space is identical to this one.
+        """
+        if not isinstance(other, DiscreteSearchSpace):
+            return NotImplemented
+        return bool(tf.reduce_all(tf.sort(self.points, 0) == tf.sort(other.points, 0)))
+
     def __deepcopy__(self, memo: dict[int, object]) -> DiscreteSearchSpace:
         return self
 
@@ -242,17 +292,17 @@ class Box(SearchSpace):
         return f"Box({self._lower!r}, {self._upper!r})"
 
     @property
-    def lower(self) -> TensorType:
+    def lower(self) -> tf.Tensor:
         """The lower bounds of the box."""
         return self._lower
 
     @property
-    def upper(self) -> TensorType:
+    def upper(self) -> tf.Tensor:
         """The upper bounds of the box."""
         return self._upper
 
     @property
-    def dimension(self) -> int:
+    def dimension(self) -> TensorType:
         """The number of inputs in this search space."""
         return self._dimension
 
@@ -279,20 +329,26 @@ class Box(SearchSpace):
 
         return tf.reduce_all(value >= self._lower) and tf.reduce_all(value <= self._upper)
 
-    def sample(self, num_samples: int) -> TensorType:
+    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
         """
         Sample randomly from the space.
 
         :param num_samples: The number of points to sample from this search space.
+        :param seed: Random seed for reproducibility.
         :return: ``num_samples`` i.i.d. random points, sampled uniformly,
             from this search space with shape '[num_samples, D]' , where D is the search space
             dimension.
         """
         tf.debugging.assert_non_negative(num_samples)
-
+        if seed is not None:  # ensure reproducibility
+            tf.random.set_seed(seed)
         dim = tf.shape(self._lower)[-1]
         return tf.random.uniform(
-            (num_samples, dim), minval=self._lower, maxval=self._upper, dtype=self._lower.dtype
+            (num_samples, dim),
+            minval=self._lower,
+            maxval=self._upper,
+            dtype=self._lower.dtype,
+            seed=seed,
         )
 
     def sample_halton(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
@@ -336,15 +392,7 @@ class Box(SearchSpace):
             dim=dim, num_results=num_samples, dtype=self._lower.dtype, skip=skip
         ) + self._lower
 
-    def discretize(self, num_samples: int) -> DiscreteSearchSpace:
-        """
-        :param num_samples: The number of points in the :class:`DiscreteSearchSpace`.
-        :return: A discrete search space consisting of ``num_samples`` points sampled uniformly from
-            this :class:`Box`.
-        """
-        return DiscreteSearchSpace(points=self.sample(num_samples))
-
-    def __mul__(self, other: Box) -> Box:
+    def product(self, other: Box) -> Box:
         r"""
         Return the Cartesian product of the two :class:`Box`\ es (concatenating their respective
         lower and upper bounds). For example:
@@ -370,6 +418,17 @@ class Box(SearchSpace):
 
         return Box(product_lower_bound, product_upper_bound)
 
+    def __eq__(self, other: object) -> bool:
+        """
+        :param other: A search space.
+        :return: Whether the search space is identical to this one.
+        """
+        if not isinstance(other, Box):
+            return NotImplemented
+        return bool(
+            tf.reduce_all(self.lower == other.lower) and tf.reduce_all(self.upper == other.upper)
+        )
+
     def __deepcopy__(self, memo: dict[int, object]) -> Box:
         return self
 
@@ -456,9 +515,9 @@ class TaggedProductSearchSpace(SearchSpace):
         return self._tags
 
     @property
-    def dimension(self) -> int:
+    def dimension(self) -> TensorType:
         """The number of inputs in this product search space."""
-        return int(self._dimension)
+        return self._dimension
 
     def get_subspace(self, tag: str) -> SearchSpace:
         """
@@ -535,7 +594,7 @@ class TaggedProductSearchSpace(SearchSpace):
         ]
         return tf.reduce_all(in_each_subspace)
 
-    def sample(self, num_samples: int) -> TensorType:
+    def sample(self, num_samples: int, seed: Optional[int] = None) -> TensorType:
         """
         Sample randomly from the space by sampling from each subspace
         and concatenating the resulting samples.
@@ -546,19 +605,12 @@ class TaggedProductSearchSpace(SearchSpace):
             dimension.
         """
         tf.debugging.assert_non_negative(num_samples)
-
-        subspace_samples = [self._spaces[tag].sample(num_samples) for tag in self._tags]
+        if seed is not None:  # ensure reproducibility
+            tf.random.set_seed(seed)
+        subspace_samples = [self._spaces[tag].sample(num_samples, seed=seed) for tag in self._tags]
         return tf.concat(subspace_samples, -1)
 
-    def discretize(self, num_samples: int) -> DiscreteSearchSpace:
-        """
-        :param num_samples: The number of points in the :class:`DiscreteSearchSpace`.
-        :return: A discrete search space consisting of ``num_samples`` points sampled
-            uniformly across the space.
-        """
-        return DiscreteSearchSpace(points=self.sample(num_samples))
-
-    def __mul__(self, other: SearchSpace) -> TaggedProductSearchSpace:
+    def product(self, other: TaggedProductSearchSpace) -> TaggedProductSearchSpace:
         r"""
         Return the Cartesian product of the two :class:`TaggedProductSearchSpace`\ s,
         building a tree of :class:`TaggedProductSearchSpace`\ s.
@@ -568,5 +620,14 @@ class TaggedProductSearchSpace(SearchSpace):
         """
         return TaggedProductSearchSpace(spaces=[self, other])
 
+    def __eq__(self, other: object) -> bool:
+        """
+        :param other: A search space.
+        :return: Whether the search space is identical to this one.
+        """
+        if not isinstance(other, TaggedProductSearchSpace):
+            return NotImplemented
+        return self._tags == other._tags and self._spaces == other._spaces
+
     def __deepcopy__(self, memo: dict[int, object]) -> TaggedProductSearchSpace:
         return self
